#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include <cv_bridge/cv_bridge.h>

#include <ros/ros.h>
#include <image_transport/image_transport.h>
#include <sensor_msgs/image_encodings.h>
#include <sensor_msgs/Image.h>
#include <geometry_msgs/PoseArray.h>
#include <geometry_msgs/Pose.h>
#include <message_filters/time_synchronizer.h>
#include <std_msgs/Int32.h>


#include <iostream>
#include <sstream>
#include <math.h>

#define PI 3.14159265

#define NO_OBJECT_DETECTION			0
#define RED_OBJECT_DETECTION		1
#define BLUE_OBJECT_DETECTION		2
//#define etc ...


using namespace std;
using namespace cv;
using namespace message_filters;
//using namespace sensor_msgs;

//static const std::string OPENCV_WINDOW = "Image window";

/**
 *  This class allow the detection of a specific type of object (color, Shape, ...)
 *  the name of redObject is used only because we can currently only detect red objects. This name must be changed for better comprehension.
 */
class ObjectDetection
{    
public:  
    ros::NodeHandle nh; /**< ROS node*/
    
    image_transport::ImageTransport it; /**< used for sending/receiving images more efficiently*/
    image_transport::Publisher pub1; /**< ROS publisher for the image contening the specific objects detected*/
    image_transport::Publisher pub2; /**< ROS publisher for the image of the mask of the original image*/
    image_transport::Subscriber subRgb; /**< ROS subsriber for the RGB image (openni_launch)*/
    image_transport::Subscriber subDepth; /**< ROS subsriber for the depth image (openni_launch)*/
    
    ros::Publisher redObjectsPositionPub; /**< ROS publisher that send */
    ros::Subscriber objectTypeSub; /**< ROS subscriber that receive the reference of the type of objects to detect*/
    
    geometry_msgs::PoseArray redObjectsPosition; /**< List of the position of the objects*/
    
    int objectType; /**< type of objects to detect*/
    
    cv_bridge::CvImagePtr cv_ptr_depth; /**< last depth image retrieved from subDepth*/
    bool isDepthReady; /**< true if there is already a depth image retrieved*/
    
    float angleKinect;
    
    /**
     * Object_Detection will subscribe and publish on the differents ROS topic. 
     * It also choose a default type of object to detect
     * 
	 * @param objectCategory type of object to initiallt detect. See #define section
     */
    ObjectDetection(int objectCategory)
	: it(nh)
    {
        isDepthReady = false;
        subRgb = it.subscribe("/camera/rgb/image_raw", 1, &ObjectDetection::rosManagement, this);
        subDepth = it.subscribe("/camera/depth/image", 1, &ObjectDetection::depthSaving, this);
        
        pub1 = it.advertise("/object_detection/red_objects", 1);
        pub2 = it.advertise("/object_detection/mask", 1);
        
        objectTypeSub = nh.subscribe("/object_detection/object_type", 10, &ObjectDetection::objectTypeCallback, this);
        redObjectsPositionPub = nh.advertise<geometry_msgs::PoseArray>("/object_detection/red_objects_position", 1000);
        
        //redObjectsPosition.poses.resize(1);
        
        objectType = objectCategory;
		angleKinect = 50 * 2*PI/360.;
    }

    ~ObjectDetection()
    {
      
    }
    
    /**
     *  Modify the type of the objects detected
     *
     *  @param msg reference to the type of objects
     */
    void objectTypeCallback(const std_msgs::Int32& msg)
    {
        objectType = msg.data;
    }
    
    /**
     *  Save the last depth image
     *
     *  @param msg depth image from kinect
     */
    void depthSaving(const sensor_msgs::ImageConstPtr& msg)
    {
      cv_ptr_depth = cv_bridge::toCvCopy(msg, sensor_msgs::image_encodings::TYPE_32FC1);
      isDepthReady = true;
    }
    
    /**
     *  Select the right type of object to detect and call the right detection function
     *  The last depth image must already be save in order to operate.
     *
     *  @param msg RGB image from kinect
     */
    
    void rosManagement(const sensor_msgs::ImageConstPtr& msg)
    {
        switch (objectType)
        {
			case NO_OBJECT_DETECTION :
			{
				break;
			}
			case RED_OBJECT_DETECTION :
			{
				cv_bridge::CvImagePtr cv_ptr;
				try
				{
					cv_ptr = cv_bridge::toCvCopy(msg, sensor_msgs::image_encodings::BGR8);
				}
				catch (cv_bridge::Exception& e)
				{
					ROS_ERROR("cv_bridge exception: %s", e.what());
					return;
				}

				Mat mask;
				if(isDepthReady)
				{
					cv_ptr->image = findColoredObject(cv_ptr->image, mask);	
					
					// Output modified video stream
					pub1.publish(cv_ptr->toImageMsg());
				}
				break;
			}          
			case BLUE_OBJECT_DETECTION :
			{
				break;
			}
        }
            
    }

    /**
     *  Find the colored objects on the picture. The color depend on the colorFiler function parameter
     *  The coordinates of the colored objects will be determined 
     *
     *  @param input RGB image
     *  @param colorOnly cv::Mat variable. It will be the mask of the image depnding of the color
     *  @return RGB image 
     */
    Mat findColoredObject(Mat& input, Mat& colorOnly)
    {
	// We filter the image in order to keep the red only
	colorOnly = colorFilter(input);

	// We filter the image in order to discard the too little red parts
	cv::Mat str_el = cv::getStructuringElement(cv::MORPH_RECT, cv::Size(3, 3));
	morphologyEx(colorOnly, colorOnly, cv::MORPH_OPEN, str_el);
	morphologyEx(colorOnly, colorOnly, cv::MORPH_CLOSE, str_el);

	//We find the contours of the red objects
	cv::vector<cv::vector<cv::Point> > contours;
	cv::vector<cv::Vec4i> hierarchy;
	cv::vector<cv::Point2f> centerObject;
	cv::vector<cv::Point2f> centerDetection;
	cv::vector<cv::RotatedRect> ellipses;
	cv::vector<int> height;
	cv::vector<int> width;

	cv::findContours( colorOnly.clone(), contours, hierarchy, CV_RETR_TREE, CV_CHAIN_APPROX_NONE);


	for( int i=0; i<contours.size(); i++)
	{
	    if (contours[i].size() < 5)
	    {
		break;
	    }
	    
	    Rect rect = boundingRect(contours[i]);
	    RotatedRect elps = fitEllipse(contours[i]);
	    
	    /*if (rect.width > 25 && rect.height > 40)
	    {
		Point2i c(rect.x + rect.width/2,rect.y + rect.height/2);
		center.push_back(c);
		width.push_back(rect.width);
		height.push_back(rect.height);
		//cout << "width : " << rect.width << " height : " << rect.height << endl;
	    }*/
	    
	    if (elps.size.width > 20 && elps.size.height > 20)
	    {
			Point2f c1(elps.center.x, elps.center.y);
			Point2f c2;
			if (elps.angle > 90)
			{
				c2.x = elps.center.x + sin((180 - elps.angle)*2*PI/360.) * elps.size.height/3.;
				c2.y = elps.center.y + cos((180 - elps.angle)*2*PI/360.) * elps.size.height/3.;
			}
			else
			{
				c2.x = elps.center.x + sin(-(elps.angle)*2*PI/360.) * elps.size.height/3.;
				c2.y = elps.center.y + cos(-(elps.angle)*2*PI/360.) * elps.size.height/3.;
			}
			if (c2.x > 0 && c2.y > 0 && c2.x < input.cols && c2.y < input.rows)
			{
				ellipses.push_back(elps);
				centerObject.push_back(c1);
				centerDetection.push_back(c2);
				width.push_back(elps.size.width);
				height.push_back(elps.size.height);
			}
	    }
	}

	cv::Scalar blue(255,0,0);
	cv::Scalar green(0,255,0);

	//cv_bridge::CvImagePtr cv_ptr_depth = cv_bridge::toCvCopy(depthImage, sensor_msgs::image_encodings::TYPE_16UC1);
	
	redObjectsPosition.poses.clear();
	bool redObjectDetected = false;
	
	for( int i = 0; i < centerDetection.size(); i++)
	{
	    //Point2i  pt1(centerDetection[i].x - width[i]/2, centerDetection[i].y + height[i]/2);
	    //Point2i  pt2(centerDetection[i].x + width[i]/2, centerDetection[i].y - height[i]/2);
	    //cv::rectangle(input,pt1,pt2, blue);	    
	    
	    float dist_val = cv_ptr_depth->image.at<float>( centerDetection[i].y, centerDetection[i].x );
	    
	    float AngularFoVHorz = 57 * 2*PI/360.;
	    
	    float degByPxHorz = AngularFoVHorz/input.cols;
	    
	    float xAngle = degByPxHorz * (centerDetection[i].x - (input.cols/2 + 0.5)); 
	    
	    float xPosition = tan(xAngle) * dist_val;
	    float xCoeff = 10./8.25;
	    float xOffset = 0;
		xPosition = floor(xPosition*xCoeff*100.)/100. - xOffset;
	    
	    float yPosition = dist_val * cos(angleKinect);
	    float yCoeff = 100./49.;
	    float yOffset = 0.32; 
		yPosition = floor(yPosition*yCoeff*100.)/100.-yOffset;
	    
	    float zPosition = sqrt(pow(dist_val,2) + pow(yPosition*yCoeff, 2));
	    float zCoeff = 1./2.35;
	    float zOffset = 0;
		zPosition = floor(zPosition*zCoeff*100.)/100.-zOffset;
	    
	    if (dist_val < 0.8 && dist_val > 0.45)
	    {
	        geometry_msgs::Pose redObjectPose;
	        redObjectPose.position.x = xPosition;
	        redObjectPose.position.y = yPosition;
	        redObjectPose.position.z = zPosition;
	        redObjectsPosition.poses.push_back(redObjectPose);
	        redObjectDetected = true;
	    
			stringstream ss1 (stringstream::in | stringstream::out);
			stringstream ss2 (stringstream::in | stringstream::out);
			
			ss1  << "(" << xPosition 
				<< "," << yPosition
				<< ") depth : " << dist_val;
				
			ss2	<< " angle : " << 180 - floor(ellipses[i].angle*100.)/100.
				<< " height : " << floor(ellipses[i].size.height*100.)/100.
				<< " width : " << floor(ellipses[i].size.width*100.)/100.;

			string dist_str1 = ss1.str();
			string dist_str2 = ss2.str();
			
			Point2f ptTxt;
			ptTxt.y = centerDetection[i].y + 15;
			ptTxt.x = centerDetection[i].x;
			cv::line(input, centerObject[i], centerDetection[i], blue);
			cv::putText(input, dist_str1, centerDetection[i], FONT_HERSHEY_SIMPLEX, 0.5, blue);
			cv::putText(input, dist_str2, ptTxt, FONT_HERSHEY_SIMPLEX, 0.3, blue);
		}
	}
	
	if(redObjectDetected)
	{
	    redObjectsPositionPub.publish(redObjectsPosition);
	}
	
	for( int i = 0; i < ellipses.size(); i++)
	{
	  cv::ellipse(input, ellipses[i], green);
	}
	
	return input;
    }
   
    /**
     *  Get the mask of the image corresponding to a range of color
     *
     *  @param input RGB image
     *  @return mask of the image 
     */   
    Mat colorFilter(const Mat& src)
    {
		assert(src.type() == CV_8UC3);

		Mat colorOnly;
		Mat red1;
		Mat red2;
		/************************************************
		*					HSV FILTER
		************************************************/
		/*cv::cvtColor(src, src, cv::COLOR_BGR2HSV);
		cv::inRange(src, cv::Scalar(0, 105, 105), cv::Scalar(10, 255, 255), red1);
		cv::inRange(src, cv::Scalar(150, 105, 105), cv::Scalar(180, 255, 255), red2);
		bitwise_or(red1, red2, colorOnly);
		cv::cvtColor(src, src, cv::COLOR_HSV2BGR);*/
		
		/************************************************
		*					BGR FILTER
		************************************************/
		//inRange(src, Scalar(0,0,90),Scalar(100,80,255), colorOnly);

		cv::inRange(src, cv::Scalar(0,0,80), cv::Scalar(70, 70, 255), red1);
		cv::inRange(src, cv::Scalar(0, 0, 120), cv::Scalar(110, 110, 255), red2);
		bitwise_or(red1, red2, colorOnly);

		return colorOnly;
    }
};

int main(int argc, char** argv)
{
    ros::init(argc, argv, "objectDetection");
    
    ObjectDetection od(RED_OBJECT_DETECTION);
    
    ros::spin();
}

